import sys
from textwrap import dedent
from threading import Thread

from common.schema import Keys
from jinja2 import Template
from PyQt6 import QtWidgets

from .config import ClientConf
from .gui.main_window import Ui_MainWindow
from .transport import JIMClient


def get_html_message_template() -> Template:
    """Возвращает шаблон HTML-разметки для рендеринга окна чата"""
    return Template(
        dedent(
            """
    <style>
        div.user {
            text-align: right;
        }
    </style>
    {% for message in messages %}
        <div class="{% if message.is_incoming %}contact{% else %}user{% endif %}">
            <p>
                <u>
                    {% if message.is_incoming %}
                        <b>{{ contact }}</b>
                    {% else %}
                        <b>{{ current_user }}</b>
                    {% endif %}
                    (<i>{{ message.time }}</i>)
                    {% if not message.is_delivered %}
                        <b>[Не доставлено]</b>
                    {% endif %}
                </u>
                <br/>
                {{ message.text }}
                <br/>
            </p>
        </div>
    {% endfor %}"""
        )
    )


class Application:
    """
    Класс, задающий поведение графического интерфейса клиента
    """

    def __init__(self) -> None:
        self.app = QtWidgets.QApplication(sys.argv)
        self.main_window = QtWidgets.QMainWindow()
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self.main_window)

        self._connect_btn_label = "Подключиться"
        self._disconnect_btn_label = "Отключиться"

        self.client: JIMClient | None = None
        self.current_chat = None
        self.prev_contact_item: QtWidgets.QListWidgetItem | None = None
        self.chat_tamplate = get_html_message_template()
        self._create_bindings()
        self._set_defaults()

    def run(self):
        """Запускает отображение интерфейса"""
        self.main_window.show()
        sys.exit(self.app.exec())

    def _set_defaults(self):
        self.ui.pushButtonConnect.setText(self._connect_btn_label)
        self.ui.listWidgetContacts.clear()
        self.ui.textBrowserChat.clear()
        self.ui.textEditMessage.clear()
        self.ui.lineEditIP.setText(ClientConf.DEFAULT_SERVER_IP)
        self.ui.lineEditPort.setText(str(ClientConf.DEFAULT_PORT))
        self.ui.textBrowserChat.setReadOnly(True)
        self._switch_controls(is_enabled=False)

    def _switch_controls(self, is_enabled: bool):
        self.ui.pushButtonAddContact.setEnabled(is_enabled)
        self.ui.pushButtonDeleteContact.setEnabled(is_enabled)
        self.ui.pushButtonSend.setEnabled(is_enabled)

    def _create_bindings(self):
        self.ui.pushButtonConnect.clicked.connect(self._on_click_connect)
        self.ui.pushButtonAddContact.clicked.connect(self._on_click_add_contact)
        self.ui.pushButtonDeleteContact.clicked.connect(self._on_click_delete_contact)
        self.ui.pushButtonSend.clicked.connect(self._on_click_send)
        self.ui.listWidgetContacts.doubleClicked.connect(self._on_contact_double_click)

    def _bind_client_signals(self):
        if self.client:
            self.client.notifier.new_message.connect(self._on_accept_new_message)
            self.client.notifier.connection_lost.connect(self._on_connection_lost)
            self.client.notifier.contacts_updated.connect(self._fill_contacts)

    def _show_standard_warning(self, info: str, title: str = "Ошибка", text: str = "", is_warning: bool = True):
        msg = QtWidgets.QMessageBox()
        msg.resize(100, 50)
        if is_warning:
            msg.setWindowIcon(self.app.style().standardIcon(QtWidgets.QStyle.StandardPixmap.SP_MessageBoxWarning))
        else:
            msg.setWindowIcon(self.app.style().standardIcon(QtWidgets.QStyle.StandardPixmap.SP_MessageBoxInformation))
        msg.setWindowTitle(title)
        msg.setText(text)
        msg.setInformativeText(info)
        msg.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok)
        msg.exec()

    def _on_click_connect(self):
        def show_empty_params_message(empty_params):
            text = "Ошибка подключения к серверу."
            info = f"Не указаны параметры: {', '.join(empty_params)}"
            self._show_standard_warning(info=info, text=text)

        if self.ui.pushButtonConnect.text() == self._connect_btn_label:
            ip = self.ui.lineEditIP.text()
            port = self.ui.lineEditPort.text()
            username = self.ui.lineEditUsername.text()
            password = self.ui.lineEditPasswd.text()
            conn_params = {
                "IP адрес": ip,
                "Порт сервера": port,
                "Имя пользователя": username,
                "Пароль": password,
            }
            if all(conn_params.values()):
                self.client = JIMClient(ip=ip, port=port, username=username, password=password)
                ok, resp = self.client.authenticate()
                if ok:
                    self._bind_client_signals()
                    self._switch_controls(is_enabled=True)
                    self.client_task = Thread(target=self.client.run)
                    self.client_task.daemon = True
                    self.client_task.start()
                    self._fill_contacts()
                    self.ui.pushButtonConnect.setText(self._disconnect_btn_label)
                elif resp:
                    message = resp.get(Keys.ALERT) or resp.get(Keys.ERROR)
                    self._show_standard_warning(info=message)  # type: ignore
                else:
                    self._show_standard_warning(info="Не удалось подключиться к серверу")  # type: ignore
            else:
                empty_params = [key for key, value in conn_params.items() if not value]
                show_empty_params_message(empty_params)
        else:
            if self.client:
                self.client.close()
            self._set_defaults()

    def _fill_contacts(self):
        current_contact_name = None
        current_contact = None
        try:
            current_contact_name = self.ui.listWidgetContacts.currentItem().text()
        except AttributeError:
            pass
        self.ui.listWidgetContacts.clear()
        if self.client:
            contacts = self.client.storage.get_contact_list()
            for contact in contacts:
                item = QtWidgets.QListWidgetItem()
                item.setText(contact)
                self.ui.listWidgetContacts.addItem(item)
                if contact == current_contact_name:
                    current_contact = item
        if current_contact:
            self.ui.listWidgetContacts.setCurrentItem(current_contact)

    def _fill_chat(self, contact_name):
        self.ui.textBrowserChat.setText(f"Пока пусто...")
        if self.client:
            messages = [
                {
                    "text": text,
                    "is_incoming": is_incoming,
                    "is_delivered": is_delivered,
                    "time": timestamp.replace(microsecond=0).isoformat().replace("T", " "),
                }
                for text, is_incoming, is_delivered, timestamp in self.client.storage.get_chat_messages(
                    contact=contact_name
                )
            ]
            if messages:
                chat_text = self._make_chat_text(
                    messages=messages, current_user=self.client.username, contact=contact_name
                )
                self.ui.textBrowserChat.setText(chat_text)
                self.ui.textBrowserChat.verticalScrollBar().setValue(
                    self.ui.textBrowserChat.verticalScrollBar().maximum()
                )

    def _make_chat_text(self, messages: list[dict], current_user: str, contact: str):
        data = {
            "messages": messages,
            "current_user": current_user,
            "contact": contact,
        }
        text = self.chat_tamplate.render(**data)
        return text

    def _on_contact_double_click(self):
        contact_item = self.ui.listWidgetContacts.currentItem()
        contact_name = contact_item.text()
        self.current_chat = contact_name
        self._fill_chat(contact_name=contact_name)

    def _on_click_send(self):
        try:
            contact = self.ui.listWidgetContacts.currentItem().text()
            if self.client and contact:
                self.client.resend_not_delivered(contact=contact)
                msg_text = self.ui.textEditMessage.toPlainText()
                if msg_text:
                    self.client.send_msg(contact=contact, msg_text=msg_text)
                    self.ui.textEditMessage.clear()
            self._fill_chat(contact_name=contact)
        except AttributeError:
            self._show_standard_warning("Выберите контакт из списка")

    def _on_click_add_contact(self):
        contact = self._show_add_contact_dialog()
        if self.client and contact:
            if contact == self.client.username:
                self._show_standard_warning(info="Нельзя добавлять себя в контакты")
                return
            ok = self.client.add_contact(contact_name=contact)
            if ok:
                self._fill_contacts()
            else:
                self._show_standard_warning(info="Контакт не найден")

    def _on_click_delete_contact(self):
        try:
            contact = self.ui.listWidgetContacts.currentItem().text()
            if contact and self.client:
                ok = self.client.delete_contact(contact_name=contact)
                if ok:
                    self._fill_contacts()
                    if self.current_chat == contact:
                        self.ui.textBrowserChat.clear()
                else:
                    self._show_standard_warning(info="Контакт не найден")
        except AttributeError:
            self._show_standard_warning("Выберите контакт из списка")

    def _show_add_contact_dialog(self):
        contact, ok = QtWidgets.QInputDialog.getText(self.main_window, "Добавление пользователя", "Введите имя:")
        if ok:
            return contact
        return None

    def _on_accept_new_message(self, sender):
        try:
            if sender == self.ui.listWidgetContacts.currentItem().text():
                self._fill_chat(sender)
        except AttributeError:
            pass
        finally:
            self._fill_contacts()

    def _on_connection_lost(self):
        self._show_standard_warning(info="Потеряно соединение с сервером.")

        self._set_defaults()
        if self.client:
            self.client.close()



from sqlalchemy import Boolean, Column, DateTime, Integer, String, Text, create_engine
from sqlalchemy.orm import declarative_base, sessionmaker

from .config import ClientConf

Base = declarative_base()


class Contact(Base):
    """
    Таблица временного хранения контактов пользователя
    """

    __tablename__ = "user"

    id = Column(Integer, primary_key=True)
    username = Column(String, index=True, unique=True)


class Message(Base):
    """
    Таблица хранения сообщений в чатах пользователя
    """

    __tablename__ = "message"

    id = Column(Integer, primary_key=True)
    contact = Column(String, index=True)
    text = Column(Text)
    timestamp = Column(DateTime)
    is_incoming = Column(Boolean, default=False, index=True)
    is_delivered = Column(Boolean, default=False, index=True)


def init_db(username: str):
    """Функция инициализации сессии ORM и базы данных при её отсутствии"""
    db_path = ClientConf.DATA_DIR / f"jim_client_db_{username}.sqlite"
    conn_string = f"sqlite:///{db_path}?check_same_thread=false&uri=true"

    engine = create_engine(conn_string, future=True, echo=False, pool_recycle=7200)
    if not db_path.exists():
        Base.metadata.create_all(engine)

    Session = sessionmaker(bind=engine)
    session = Session()
    return session, db_path



from datetime import datetime

from sqlalchemy import and_
from sqlalchemy.exc import NoResultFound

from .model import Contact, Message, init_db


class ClientStorage:
    """
    Класс-обертка над ORM для взаимодействия клиента с базой данных
    """

    def __init__(self, username) -> None:
        session, db_path = init_db(username)
        self.session = session
        self.db_path = db_path

    def get_contact_list(self):
        query = self.session.query(Contact).all()
        return [(entry.username) for entry in query]

    def check_contact(self, contact: str):
        if self.session.query(Contact).filter_by(username=contact).count():
            return True
        else:
            return False

    def clear_contacts(self):
        self.session.query(Contact).delete()
        self.session.commit()

    def update_contacts(self, contacts: list):
        self.clear_contacts()
        for new_contact in contacts:
            self.add_contact(new_contact)

    def store_msg(self, contact: str, msg_text: str, timestamp: datetime, is_incoming: bool, is_delivered: bool = True):
        message = Message(
            contact=contact, text=msg_text, timestamp=timestamp, is_incoming=is_incoming, is_delivered=is_delivered
        )
        self.session.add(message)
        self.session.commit()

    def get_chat_messages(self, contact):
        try:
            messages = self.session.query(Message).filter_by(contact=contact).all()
        except NoResultFound:
            return []
        else:
            return [(entry.text, entry.is_incoming, entry.is_delivered, entry.timestamp) for entry in messages]

    def get_not_delivered_msgs(self, contact: str):
        try:
            messages = (
                self.session.query(Message)
                .filter(
                    and_(
                        Message.contact == contact,
                        Message.is_incoming == False,
                        Message.is_delivered == False,
                    )
                )
                .all()
            )
        except NoResultFound:
            return []
        else:
            return [(entry.id, entry.text) for entry in messages]

    def mark_msg_delivered(self, msg_id: int):
        try:
            msg = self.session.query(Message).filter_by(id=msg_id).one()
            msg.is_delivered = True
            self.session.commit()
        except NoResultFound:
            pass

    def add_contact(self, contact: str):
        try:
            self.session.query(Contact).filter_by(username=contact).one()
        except NoResultFound:
            contact_obj = Contact(username=contact)
            self.session.add(contact_obj)
            self.session.commit()

    def del_contact(self, contact: str):
        try:
            self.session.query(Contact).filter_by(username=contact).delete()
            self.session.commit()
        except NoResultFound:
            pass



# Form implementation generated from reading ui file './client/gui/client_gui.ui'
#
# Created by: PyQt6 UI code generator 6.4.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1000, 800)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.centralwidget)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.widgetMain = QtWidgets.QWidget(self.centralwidget)
        self.widgetMain.setObjectName("widgetMain")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.widgetMain)
        self.verticalLayout.setObjectName("verticalLayout")
        self.groupBoxConnection = QtWidgets.QGroupBox(self.widgetMain)
        self.groupBoxConnection.setObjectName("groupBoxConnection")
        self.gridLayout = QtWidgets.QGridLayout(self.groupBoxConnection)
        self.gridLayout.setObjectName("gridLayout")
        self.lineEditIP = QtWidgets.QLineEdit(self.groupBoxConnection)
        self.lineEditIP.setObjectName("lineEditIP")
        self.gridLayout.addWidget(self.lineEditIP, 1, 0, 1, 1)
        self.labelIP = QtWidgets.QLabel(self.groupBoxConnection)
        self.labelIP.setObjectName("labelIP")
        self.gridLayout.addWidget(self.labelIP, 0, 0, 1, 1)
        self.pushButtonConnect = QtWidgets.QPushButton(self.groupBoxConnection)
        self.pushButtonConnect.setObjectName("pushButtonConnect")
        self.gridLayout.addWidget(self.pushButtonConnect, 1, 4, 1, 1)
        self.labelPort = QtWidgets.QLabel(self.groupBoxConnection)
        self.labelPort.setObjectName("labelPort")
        self.gridLayout.addWidget(self.labelPort, 0, 1, 1, 1)
        self.lineEditUsername = QtWidgets.QLineEdit(self.groupBoxConnection)
        self.lineEditUsername.setObjectName("lineEditUsername")
        self.gridLayout.addWidget(self.lineEditUsername, 1, 2, 1, 1)
        self.labelUsername = QtWidgets.QLabel(self.groupBoxConnection)
        self.labelUsername.setObjectName("labelUsername")
        self.gridLayout.addWidget(self.labelUsername, 0, 2, 1, 1)
        self.lineEditPort = QtWidgets.QLineEdit(self.groupBoxConnection)
        self.lineEditPort.setObjectName("lineEditPort")
        self.gridLayout.addWidget(self.lineEditPort, 1, 1, 1, 1)
        self.lineEditPasswd = QtWidgets.QLineEdit(self.groupBoxConnection)
        self.lineEditPasswd.setEchoMode(QtWidgets.QLineEdit.EchoMode.Password)
        self.lineEditPasswd.setObjectName("lineEditPasswd")
        self.gridLayout.addWidget(self.lineEditPasswd, 1, 3, 1, 1)
        self.labelPasswd = QtWidgets.QLabel(self.groupBoxConnection)
        self.labelPasswd.setObjectName("labelPasswd")
        self.gridLayout.addWidget(self.labelPasswd, 0, 3, 1, 1)
        self.verticalLayout.addWidget(self.groupBoxConnection)
        self.groupBoxChat = QtWidgets.QGroupBox(self.widgetMain)
        self.groupBoxChat.setObjectName("groupBoxChat")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.groupBoxChat)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.textBrowserChat = QtWidgets.QTextBrowser(self.groupBoxChat)
        self.textBrowserChat.setObjectName("textBrowserChat")
        self.verticalLayout_4.addWidget(self.textBrowserChat)
        self.textEditMessage = QtWidgets.QTextEdit(self.groupBoxChat)
        self.textEditMessage.setMaximumSize(QtCore.QSize(16777215, 150))
        self.textEditMessage.setObjectName("textEditMessage")
        self.verticalLayout_4.addWidget(self.textEditMessage)
        self.pushButtonSend = QtWidgets.QPushButton(self.groupBoxChat)
        self.pushButtonSend.setObjectName("pushButtonSend")
        self.verticalLayout_4.addWidget(self.pushButtonSend)
        self.verticalLayout.addWidget(self.groupBoxChat)
        self.horizontalLayout.addWidget(self.widgetMain)
        self.widgetContacts = QtWidgets.QWidget(self.centralwidget)
        self.widgetContacts.setMaximumSize(QtCore.QSize(300, 16777215))
        self.widgetContacts.setObjectName("widgetContacts")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.widgetContacts)
        self.verticalLayout_3.setSizeConstraint(QtWidgets.QLayout.SizeConstraint.SetDefaultConstraint)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.groupBoxContacts = QtWidgets.QGroupBox(self.widgetContacts)
        self.groupBoxContacts.setObjectName("groupBoxContacts")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBoxContacts)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.listWidgetContacts = QtWidgets.QListWidget(self.groupBoxContacts)
        self.listWidgetContacts.setObjectName("listWidgetContacts")
        self.verticalLayout_2.addWidget(self.listWidgetContacts)
        self.pushButtonAddContact = QtWidgets.QPushButton(self.groupBoxContacts)
        self.pushButtonAddContact.setObjectName("pushButtonAddContact")
        self.verticalLayout_2.addWidget(self.pushButtonAddContact)
        self.pushButtonDeleteContact = QtWidgets.QPushButton(self.groupBoxContacts)
        self.pushButtonDeleteContact.setObjectName("pushButtonDeleteContact")
        self.verticalLayout_2.addWidget(self.pushButtonDeleteContact)
        self.verticalLayout_3.addWidget(self.groupBoxContacts)
        self.horizontalLayout.addWidget(self.widgetContacts)
        MainWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)
        MainWindow.setTabOrder(self.lineEditUsername, self.lineEditPasswd)
        MainWindow.setTabOrder(self.lineEditPasswd, self.pushButtonConnect)
        MainWindow.setTabOrder(self.pushButtonConnect, self.listWidgetContacts)
        MainWindow.setTabOrder(self.listWidgetContacts, self.textEditMessage)
        MainWindow.setTabOrder(self.textEditMessage, self.pushButtonSend)
        MainWindow.setTabOrder(self.pushButtonSend, self.pushButtonAddContact)
        MainWindow.setTabOrder(self.pushButtonAddContact, self.pushButtonDeleteContact)
        MainWindow.setTabOrder(self.pushButtonDeleteContact, self.textBrowserChat)
        MainWindow.setTabOrder(self.textBrowserChat, self.lineEditIP)
        MainWindow.setTabOrder(self.lineEditIP, self.lineEditPort)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "GB PyChat"))
        self.groupBoxConnection.setTitle(_translate("MainWindow", "Настройки подключения"))
        self.lineEditIP.setPlaceholderText(_translate("MainWindow", "127.0.0.1"))
        self.labelIP.setText(_translate("MainWindow", "IP адрес сервера:"))
        self.pushButtonConnect.setText(_translate("MainWindow", "Подключиться"))
        self.labelPort.setText(_translate("MainWindow", "Порт сервера:"))
        self.labelUsername.setText(_translate("MainWindow", "Имя пользователя:"))
        self.lineEditPort.setPlaceholderText(_translate("MainWindow", "7777"))
        self.labelPasswd.setText(_translate("MainWindow", "Пароль:"))
        self.groupBoxChat.setTitle(_translate("MainWindow", "Чат с пользователем"))
        self.pushButtonSend.setText(_translate("MainWindow", "Отправить"))
        self.groupBoxContacts.setTitle(_translate("MainWindow", "Контакты"))
        self.pushButtonAddContact.setText(_translate("MainWindow", "Добавить контакт"))
        self.pushButtonDeleteContact.setText(_translate("MainWindow", "Удалить контакт"))


if __name__ == "__main__":
    import sys

    app = QtWidgets.QApplication(sys.argv)
    MainWindow = QtWidgets.QMainWindow()
    ui = Ui_MainWindow()
    ui.setupUi(MainWindow)
    MainWindow.show()
    sys.exit(app.exec())



from contextlib import ContextDecorator
from http import HTTPStatus
from ipaddress import ip_address
from socket import AF_INET, SOCK_STREAM, socket
from threading import Lock, Thread
from time import sleep

from common.base import JIMBase
from common.descriptors import PortDescriptor
from common.errors import IncorrectDataRecivedError, NonDictInputError, ReqiuredFieldMissingError, ServerDisconnectError
from common.schema import Actions, Keys
from PyQt6 import QtCore

from .logger_conf import main_logger
from .messages import ClientMessages
from .storage import ClientStorage

socket_lock = Lock()


class SignalNotifier(QtCore.QObject):
    """
    Класс-генератор сигналов PyQt для взаимодействия с графическим интерфейсом
    """

    new_message = QtCore.pyqtSignal(str)
    connection_lost = QtCore.pyqtSignal()
    contacts_updated = QtCore.pyqtSignal()

    def __init__(self) -> None:
        super().__init__()


class JIMClient(Thread, JIMBase, ContextDecorator):
    """
    Класс ядра клиента, отвечает за прием и передачу сообщений, а также взаимодействие с базой данных
    """

    port = PortDescriptor()

    def __init__(self, ip: str, port: int | str, username: str, password: str) -> None:
        Thread.__init__(self)
        JIMBase.__init__(self)
        self.connected = False
        self.ip = ip_address(ip)
        self.port = port
        self.server_name = f"{self.ip}:{self.port}"
        self.sock = socket(AF_INET, SOCK_STREAM)
        self.username = username
        self.password = password
        self.msg_factory = ClientMessages(self.username, self.encoding)
        self.storage = ClientStorage(self.username)
        self.notifier = SignalNotifier()
        self.status = ""

    def __str__(self):
        return f"JIM_client_object"

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def _connect(self):
        while True:
            try:
                self.sock.connect((str(self.ip), self.port))
                self.connected = True
                break
            except ConnectionRefusedError as ex:
                main_logger.info(f"Пытаюсь подключиться как {self.username} к серверу {self.server_name}...")
                sleep(1)
            except (NonDictInputError, IncorrectDataRecivedError, ReqiuredFieldMissingError) as ex:
                main_logger.info(
                    f"Не удалось подключиться от имени {self.username} к серверу {self.server_name} ({str(ex)})"
                )
                break

    def close(self):
        main_logger.info("Закрываю соединение...")
        self.sock.close()

    def authenticate(self):
        """Запускает подключение и аутентификацию клиента на сервере"""
        self._connect()
        if self.connected:
            msg = self.msg_factory.make_authenticate_msg(password=self.password)
            resp = self._send_data(msg, return_response=True)
            if resp.get(Keys.RESPONSE) == HTTPStatus.OK:  # type: ignore
                main_logger.info(f"Успешно подключен к серверу {self.server_name} от имени {self.username}")
                self.send_presence(status=self.status)
                return True, resp
            return False, resp
        return False, dict()

    def run(self):
        """Запускает синхронизацию контактов и фоновый поток обработки сообщений после подключения и аутентификации"""
        if self.connected:
            self.sync_contacts()
            receiver = Thread(target=self._start_reciever_loop)
            receiver.daemon = True
            receiver.start()
            while True:
                sleep(1)
                if receiver.is_alive():
                    continue
                break

    def sync_contacts(self):
        msg = self.msg_factory.make_get_contacts_msg()
        resp = self._send_data(msg, return_response=True)
        if resp[Keys.RESPONSE] == HTTPStatus.ACCEPTED:  # type: ignore
            contacts = resp.get(Keys.ALERT)  # type: ignore
            self.storage.update_contacts(contacts)  # type: ignore
            self.notifier.contacts_updated.emit()

    def send_presence(self, status: str):
        msg = self.msg_factory.make_presence_msg(status=status)
        self._send_data(msg)

    def send_msg(self, contact: str, msg_text: str, store_msg: bool = True):
        msg = self.msg_factory.make_msg(user_or_room=contact, message=msg_text)
        timestamp = self._from_iso_to_datetime(msg[Keys.TIME])
        resp = self._send_data(msg, return_response=True)
        is_delivered = False
        if resp and resp.get(Keys.RESPONSE) == HTTPStatus.OK:
            is_delivered = True
        if store_msg:
            self.storage.store_msg(
                contact=contact, msg_text=msg_text, timestamp=timestamp, is_incoming=False, is_delivered=is_delivered
            )
        return is_delivered

    def resend_not_delivered(self, contact: str):
        messages = self.storage.get_not_delivered_msgs(contact=contact)
        for msg_id, msg_text in messages:
            is_delivered = self.send_msg(contact=contact, msg_text=msg_text, store_msg=False)
            if is_delivered:
                self.storage.mark_msg_delivered(msg_id=msg_id)
            sleep(0.1)

    def add_contact(self, contact_name):
        msg = self.msg_factory.make_add_contact_msg(contact=contact_name)
        resp = self._send_data(msg, return_response=True)
        if resp and resp.get(Keys.RESPONSE) == HTTPStatus.OK:
            self.storage.add_contact(contact=contact_name)
            return True
        return False

    def delete_contact(self, contact_name):
        msg = self.msg_factory.make_del_contact_msg(contact=contact_name)
        resp = self._send_data(msg, return_response=True)
        if resp and resp.get(Keys.RESPONSE) == HTTPStatus.OK:
            self.storage.del_contact(contact=contact_name)
            return True
        return False

    def _start_reciever_loop(self):
        while True:
            sleep(1)
            with socket_lock:
                self.sock.settimeout(0.5)
                try:
                    msg = self._recv()
                    try:
                        self._validate_msg(msg)
                    except (NonDictInputError, IncorrectDataRecivedError, ReqiuredFieldMissingError) as ex:
                        main_logger.error(f"Принято некорректное сообщение: {msg} ({ex})")
                    else:
                        main_logger.info(f"Принято сообщение: {msg}")
                        self._process_server_msg(msg)
                except (ConnectionError, ConnectionAbortedError, ConnectionResetError, ServerDisconnectError):
                    main_logger.info("Потеряно соединение с сервером.")
                    self.notifier.connection_lost.emit()
                    break
                except OSError as ex:
                    if ex.errno:
                        main_logger.info("Потеряно соединение с сервером.")
                        self.notifier.connection_lost.emit()
                        break
                finally:
                    self.sock.settimeout(5)

    def _process_server_msg(self, msg):
        if msg.get(Keys.ACTION) == Actions.MSG:
            user_from = msg[Keys.FROM]
            if not self.storage.check_contact(contact=user_from):
                self.storage.add_contact(contact=user_from)
            text = msg[Keys.MSG]
            timestamp = self._from_iso_to_datetime(msg[Keys.TIME])
            self.storage.store_msg(contact=user_from, msg_text=text, is_incoming=True, timestamp=timestamp)
            self.notifier.new_message.emit(user_from)
            main_logger.debug(f"Получено сообщение: {msg}")
        elif msg.get(Keys.ACTION) == Actions.PROBE:
            self.send_presence(status=self.status)
            main_logger.debug(f"Получено сообщение от сервера: {msg}")
        else:
            main_logger.error(f"Сообщение не распознано: {msg}")

    def _send_data(self, msg_data: dict, return_response: bool = False):
        with socket_lock:
            msg_raw_data = self._dump_msg(msg_data)
            self.sock.send(msg_raw_data)
            resp = self._recv()
        main_logger.debug(f"Отправлено сообщение: {msg_data}")
        main_logger.debug(f"Принят ответ: {resp}")
        if return_response:
            return resp

    def _recv(self) -> dict:
        raw_resp_data = self.sock.recv(self.package_length)
        if not raw_resp_data:
            raise ServerDisconnectError()
        return self._load_msg(raw_resp_data)



from datetime import datetime

from common.schema import Actions, Keys


class ClientMessages:
    """
    Класс для генерации валидных сообщений клиента по протоколу JIM
    """

    def __init__(self, username: str, encoding: str) -> None:
        self.username = username
        self.encoding = encoding

    def _update_timestamp(self, msg: dict):
        timestamp = {Keys.TIME: datetime.now().isoformat()}
        msg.update(timestamp)

    def make_presence_msg(self, status: str = ""):
        msg = {Keys.ACTION: Actions.PRESENCE, Keys.USER: {Keys.ACCOUNT_NAME: self.username, Keys.STATUS: status}}
        self._update_timestamp(msg=msg)
        return msg

    def make_authenticate_msg(self, password: str):
        msg = {
            Keys.ACTION: Actions.AUTH,
            Keys.USER: {Keys.ACCOUNT_NAME: self.username, Keys.PASSWORD: password},
        }
        self._update_timestamp(msg=msg)
        return msg

    def make_quit_msg(self):
        msg = {
            Keys.ACTION: Actions.QUIT,
        }
        self._update_timestamp(msg=msg)
        return msg

    def make_msg(self, user_or_room: str, message: str):
        msg = {
            Keys.ACTION: Actions.MSG,
            Keys.FROM: self.username,
            Keys.TO: user_or_room,
            Keys.MSG: message,
            Keys.ENCODING: self.encoding,
        }
        self._update_timestamp(msg=msg)
        return msg

    def make_join_room_msg(self, room_name: str):
        room_name = room_name if room_name.startswith("#") else f"#{room_name}"
        msg = {Keys.ACTION: Actions.JOIN, Keys.ROOM: room_name}
        self._update_timestamp(msg=msg)
        return msg

    def make_leave_room_msg(self, room_name: str):
        room_name = room_name if room_name.startswith("#") else f"#{room_name}"
        msg = {Keys.ACTION: Actions.LEAVE, Keys.ROOM: room_name}
        self._update_timestamp(msg=msg)
        return msg

    def make_get_contacts_msg(self):
        msg = {Keys.ACTION: Actions.CONTACTS, Keys.ACCOUNT_NAME: self.username}
        self._update_timestamp(msg=msg)
        return msg

    def make_add_contact_msg(self, contact: str):
        msg = {Keys.ACTION: Actions.ADD_CONTACT, Keys.ACCOUNT_NAME: self.username, Keys.CONTACT: contact}
        self._update_timestamp(msg=msg)
        return msg

    def make_del_contact_msg(self, contact: str):
        msg = {Keys.ACTION: Actions.DEL_CONTACT, Keys.ACCOUNT_NAME: self.username, Keys.CONTACT: contact}
        self._update_timestamp(msg=msg)
        return msg



import os
import pathlib

from common.config import CommonConf


class ClientConf(CommonConf):
    """
    Настройки клиента по умолчанию
    """

    ROOT_DIR = pathlib.Path().resolve()
    LOGS_DIR = ROOT_DIR / "logs"
    DATA_DIR = ROOT_DIR / "data"
    DEFAULT_SERVER_IP = "127.0.0.1"
    MAIN_LOG_FILE_PATH = LOGS_DIR / "client.error.log"
    CALL_LOG_FILE_PATH = LOGS_DIR / "client.calls.log"


os.makedirs(ClientConf.LOGS_DIR, exist_ok=True)
os.makedirs(ClientConf.DATA_DIR, exist_ok=True)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['client', 'client.gui']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'PyQt6>=6.4.0,<7.0.0',
 'SQLAlchemy>=1.4.44,<2.0.0',
 'gb-pychat-common @ ../common']

entry_points = \
{'console_scripts': ['client-gui = client.run_client:run']}

setup_kwargs = {
    'name': 'gb-pychat-client',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'DashViolin',
    'author_email': 'mymail@srv.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
